﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Markup;
using System.Windows.Media;

namespace FeedbackAndSmoothZ
{
    [ContentProperty("Child")]
    public partial class ManipulablePictureFrame : UserControl
    {
        public static readonly DependencyProperty ChildProperty =
            DependencyProperty.Register("Child",
                typeof(UIElement),
                typeof(ManipulablePictureFrame),
                new PropertyMetadata(OnChildChanged));

        static ManipulablePictureFrame()
        {
            // Re-set defaults on three properties
            HorizontalAlignmentProperty.OverrideMetadata(typeof(ManipulablePictureFrame),
                            new FrameworkPropertyMetadata(HorizontalAlignment.Left));

            VerticalAlignmentProperty.OverrideMetadata(typeof(ManipulablePictureFrame),
                            new FrameworkPropertyMetadata(VerticalAlignment.Top));

            IsManipulationEnabledProperty.OverrideMetadata(typeof(ManipulablePictureFrame),
                            new FrameworkPropertyMetadata(true));
        }

        public ManipulablePictureFrame()
        {
            InitializeComponent();

            // If there's not a MatrixTransform available, create one
            if (RenderTransform != null)
            {
                if (!(RenderTransform is MatrixTransform))
                    RenderTransform = new MatrixTransform(RenderTransform.Value);
            }
            else
            {
                RenderTransform = new MatrixTransform();
            }
        }

        public UIElement Child
        {
            set { SetValue(ChildProperty, value); }
            get { return (UIElement)GetValue(ChildProperty); }
        }

        static void OnChildChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as ManipulablePictureFrame).border.Child = args.NewValue as UIElement;
        }

        protected override void OnManipulationStarting(ManipulationStartingEventArgs args)
        {
            VisualStateManager.GoToElementState(this, "Touched", false);

            args.ManipulationContainer = this.Parent as UIElement;
            args.Handled = true;
            base.OnManipulationStarting(args);
        }

        protected override void OnManipulationDelta(ManipulationDeltaEventArgs args)
        {
            MatrixTransform xform = this.RenderTransform as MatrixTransform;
            Matrix matrix = xform.Matrix;
            ManipulationDelta delta = args.DeltaManipulation;
            Point center = args.ManipulationOrigin;
            matrix.Translate(-center.X, -center.Y);
            matrix.Scale(delta.Scale.X, delta.Scale.Y);
            matrix.Rotate(delta.Rotation);
            matrix.Translate(center.X, center.Y);
            matrix.Translate(delta.Translation.X, delta.Translation.Y);
            xform.Matrix = matrix;

            BumpUpZIndex(this, (this.Parent as Panel).Children);

            // Rotate drop shadow
            dropShadow.Direction += delta.Rotation;

            args.Handled = true;
            base.OnManipulationDelta(args);
        }

        protected override void OnManipulationCompleted(ManipulationCompletedEventArgs args)
        {
            VisualStateManager.GoToElementState(this, "Untouched", false);
            base.OnManipulationCompleted(args);
        }

        // BumpUpZIndex with re-usable SortedDictionary object
        SortedDictionary<int, UIElement> childrenByZIndex = new SortedDictionary<int, UIElement>();

        void BumpUpZIndex(FrameworkElement touchedElement, UIElementCollection siblings)
        {
            // Make sure everybody has a unique even ZIndex
            for (int childIndex = 0; childIndex < siblings.Count; childIndex++)
            {
                UIElement child = siblings[childIndex];
                int zIndex = Panel.GetZIndex(child);
                Panel.SetZIndex(child, 2 * (zIndex * siblings.Count + childIndex));
            }

            int zIndexNew = Panel.GetZIndex(touchedElement);
            int zIndexCantGoBeyond = Int32.MaxValue;

            // Don't want to jump ahead of any intersecting elements that are on top
            foreach (UIElement child in siblings)
                if (child != touchedElement && 
                    AreElementsIntersecting(touchedElement, (FrameworkElement)child))
                {
                    int zIndexChild = Panel.GetZIndex(child);

                    if (zIndexChild > Panel.GetZIndex(touchedElement))
                        zIndexCantGoBeyond = Math.Min(zIndexCantGoBeyond, zIndexChild);
                }

            // But want to be in front of non-intersecting elements
            foreach (UIElement child in siblings)
                if (child != touchedElement && 
                    !AreElementsIntersecting(touchedElement, (FrameworkElement)child))
                {
                    // This ZIndex is odd, hence unique
                    int zIndexNextHigher = 1 + Panel.GetZIndex(child);

                    if (zIndexNextHigher < zIndexCantGoBeyond)
                        zIndexNew = Math.Max(zIndexNew, zIndexNextHigher);
                }

            // Now give all elements indices from 0 to (siblings.Count - 1)
            Panel.SetZIndex(touchedElement, zIndexNew);
            childrenByZIndex.Clear();
            int index = 0;

            foreach (UIElement child in siblings)
                childrenByZIndex.Add(Panel.GetZIndex(child), child);

            foreach (UIElement child in childrenByZIndex.Values)
                Panel.SetZIndex(child, index++);
            }

        // Test if elements are intersecting with re-usable RectangleGeometry objects
        RectangleGeometry rectGeo1 = new RectangleGeometry();
        RectangleGeometry rectGeo2 = new RectangleGeometry();

        bool AreElementsIntersecting(FrameworkElement element1, FrameworkElement element2)
        {
            rectGeo1.Rect = new Rect(new Size(element1.ActualWidth, element1.ActualHeight));
            rectGeo1.Transform = element1.RenderTransform;

            rectGeo2.Rect = new Rect(new Size(element2.ActualWidth, element2.ActualHeight));
            rectGeo2.Transform = element2.RenderTransform;

            return rectGeo1.FillContainsWithDetail(rectGeo2) != IntersectionDetail.Empty;
        }
    }
}
