﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Markup;

namespace NoTransformManipulation
{
    [ContentProperty("Child")]
    public partial class NoTransformPictureFrame : UserControl
    {
        public static readonly DependencyProperty ChildProperty =
            DependencyProperty.Register("Child",
                typeof(UIElement),
                typeof(NoTransformPictureFrame),
                new PropertyMetadata(OnChildChanged));

        static NoTransformPictureFrame()
        {
            // Re-set default on property
            IsManipulationEnabledProperty.OverrideMetadata(typeof(NoTransformPictureFrame),
                            new FrameworkPropertyMetadata(true));
        }

        public NoTransformPictureFrame()
        {
            InitializeComponent();
        }

        public UIElement Child
        {
            set { SetValue(ChildProperty, value); }
            get { return (UIElement)GetValue(ChildProperty); }
        }

        static void OnChildChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as NoTransformPictureFrame).border.Child = args.NewValue as UIElement;
        }

        protected override void OnManipulationStarting(ManipulationStartingEventArgs args)
        {
            VisualStateManager.GoToElementState(this, "Touched", false);

            // Don't allow rotation!
            args.Mode = ManipulationModes.All & ~ManipulationModes.Rotate;

            args.ManipulationContainer = this.Parent as UIElement;
            args.Handled = true;
            base.OnManipulationStarting(args);
        }

        protected override void OnManipulationDelta(ManipulationDeltaEventArgs args)
        {
            ManipulationDelta delta = args.DeltaManipulation;

            // Perform translation
            double x = Canvas.GetLeft(this) + delta.Translation.X;
            double y = Canvas.GetTop(this) + delta.Translation.Y;

            // Adjust for manipulation origin
            double fractionX = (args.ManipulationOrigin.X - x) / this.ActualWidth;
            double fractionY = (args.ManipulationOrigin.Y - y) / this.ActualHeight;

            Canvas.SetLeft(this, x - fractionX * delta.Expansion.X);
            Canvas.SetTop(this, y - fractionY * delta.Expansion.Y);

            // Perform dimension change
            this.Width = this.ActualWidth + delta.Expansion.X;

            BumpUpZIndex(this, (this.Parent as Panel).Children);

            args.Handled = true;
            base.OnManipulationDelta(args);
        }

        protected override void OnManipulationCompleted(ManipulationCompletedEventArgs args)
        {
            VisualStateManager.GoToElementState(this, "Untouched", false);
            base.OnManipulationCompleted(args);
        }

        // BumpUpZIndex with re-usable SortedDictionary object
        SortedDictionary<int, UIElement> childrenByZIndex = new SortedDictionary<int, UIElement>();

        void BumpUpZIndex(FrameworkElement touchedElement, UIElementCollection siblings)
        {
            // Make sure everybody has a unique even ZIndex
            for (int childIndex = 0; childIndex < siblings.Count; childIndex++)
            {
                UIElement child = siblings[childIndex];
                int zIndex = Panel.GetZIndex(child);
                Panel.SetZIndex(child, 2 * (zIndex * siblings.Count + childIndex));
            }

            int zIndexNew = Panel.GetZIndex(touchedElement);
            int zIndexCantGoBeyond = Int32.MaxValue;

            // Don't want to jump ahead of any intersecting elements that are on top
            foreach (UIElement child in siblings)
                if (child != touchedElement &&
                    AreElementsIntersecting(touchedElement, (FrameworkElement)child))
                {
                    int zIndexChild = Panel.GetZIndex(child);

                    if (zIndexChild > Panel.GetZIndex(touchedElement))
                        zIndexCantGoBeyond = Math.Min(zIndexCantGoBeyond, zIndexChild);
                }

            // But want to be in front of non-intersecting elements
            foreach (UIElement child in siblings)
                if (child != touchedElement &&
                    !AreElementsIntersecting(touchedElement, (FrameworkElement)child))
                {
                    // This ZIndex is odd, hence unique
                    int zIndexNextHigher = 1 + Panel.GetZIndex(child);

                    if (zIndexNextHigher < zIndexCantGoBeyond)
                        zIndexNew = Math.Max(zIndexNew, zIndexNextHigher);
                }

            // Now give all elements indices from 0 to (siblings.Count - 1)
            Panel.SetZIndex(touchedElement, zIndexNew);
            childrenByZIndex.Clear();
            int index = 0;

            foreach (UIElement child in siblings)
                childrenByZIndex.Add(Panel.GetZIndex(child), child);

            foreach (UIElement child in childrenByZIndex.Values)
                Panel.SetZIndex(child, index++);
        }

        // Test if elements are intersecting with re-usable RectangleGeometry objects
        RectangleGeometry rectGeo1 = new RectangleGeometry();
        RectangleGeometry rectGeo2 = new RectangleGeometry();

        bool AreElementsIntersecting(FrameworkElement element1, FrameworkElement element2)
        {
            rectGeo1.Rect = new Rect(Canvas.GetLeft(element1), Canvas.GetTop(element1), 
                                        element1.ActualWidth, element1.ActualHeight);

            rectGeo2.Rect = new Rect(Canvas.GetLeft(element2), Canvas.GetTop(element2), 
                                        element2.ActualWidth, element2.ActualHeight);

            return rectGeo1.FillContainsWithDetail(rectGeo2) != IntersectionDetail.Empty;
        }
    }
}
